@@ -5,6 +5,10 @@
  * is to reset password based information.
  * There is also a simple commandline based registry editor included.
  * 
+ * 2008-mar: Minor other tweaks
+ * 2008-mar: Interactive reg ed moved out of this file, into edlib.c
+ * 2008-mar: 64 bit compatible patch by Mike Doty, via Alon Bar-Lev
+ *           http://bugs.gentoo.org/show_bug.cgi?id=185411
  * 2007-sep: Group handling extended, promotion now public
  * 2007-sep: User edit menu, some changes to user info edit
  * 2007-apr-may: Get and display users group memberships
@@ -25,7 +29,7 @@
  *
  *****
  *
- * Copyright (c) 1997-2007 Petter Nordahl-Hagen.
+ * Copyright (c) 1997-2008 Petter Nordahl-Hagen.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -54,6 +58,8 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <sys/types.h>
+#include <inttypes.h>
 
 #include <openssl/des.h>
 #include <openssl/md4.h>
@@ -65,7 +71,7 @@
 #include "ntreg.h"
 #include "sam.h"
 
-const char chntpw_version[] = "chntpw version 0.99.5 070923 (decade), (c) Petter N Hagen";
+const char chntpw_version[] = "chntpw version 0.99.6 080526 (sixtyfour), (c) Petter N Hagen";
 
 extern char *val_types[REG_MAX+1];
 
@@ -139,7 +145,7 @@ void str_to_key(unsigned char *str,unsigned char *key)
  * Function to convert the RID to the first decrypt key.
  */
 
-void sid_to_key1(unsigned long sid,unsigned char deskey[8])
+void sid_to_key1(uint32_t sid,unsigned char deskey[8])
 {
 	unsigned char s[7];
 
@@ -158,7 +164,7 @@ void sid_to_key1(unsigned long sid,unsigned char deskey[8])
  * Function to convert the RID to the second decrypt key.
  */
 
-void sid_to_key2(unsigned long sid,unsigned char deskey[8])
+void sid_to_key2(uint32_t sid,unsigned char deskey[8])
 {
 	unsigned char s[7];
 	
@@ -202,7 +208,7 @@ void check_get_samdata(void)
   if (H_SAM >= 0) {
 
     /* Get users F value */
-    v = get_val2buf(hive[H_SAM], NULL, 0, ACCOUNTDB_F_PATH, REG_BINARY);
+    v = get_val2buf(hive[H_SAM], NULL, 0, ACCOUNTDB_F_PATH, REG_BINARY, TPF_VK);
     if (!v) {
       printf("Login counts data not found in SAM\n");
       return;
@@ -240,7 +246,7 @@ short handle_F(int rid, int mode)
 
   /* Get users F value */
   snprintf(s,180,"\\SAM\\Domains\\Account\\Users\\%08X\\F",rid);
-  v = get_val2buf(hive[H_SAM], NULL, 0, s, REG_BINARY);
+  v = get_val2buf(hive[H_SAM], NULL, 0, s, REG_BINARY, TPF_VK_EXACT);
   if (!v) {
     printf("Cannot find value <%s>\n",s);
     return(0);
@@ -275,7 +281,7 @@ short handle_F(int rid, int mode)
     acb &= ~ACB_AUTOLOCK;
     f->ACB_bits = acb;
     f->failedcnt = 0;
-    put_buf2val(hive[H_SAM], v, 0, s, REG_BINARY);
+    put_buf2val(hive[H_SAM], v, 0, s, REG_BINARY,TPF_VK_EXACT);
     printf("Unlocked!\n");
   }
   return (acb | ( (f->failedcnt > 0 && f->failedcnt >= max_sam_lock)<<15 ) | (acb & ACB_AUTOLOCK)<<15 | (acb & ACB_DISABLED)<<15);
@@ -315,7 +321,7 @@ int list_user_groups(int rid, int check)
     return(0);
   }
   nk += 4;
-  count = get_val_type(hive[H_SAM],nk,"@");
+  count = get_val_type(hive[H_SAM],nk,"@",TPF_VK_EXACT);
   if (count == -1) {
     printf("list_user_groups(): Cannot find value <%s\\@>\n",s);
     return(0);
@@ -324,7 +330,7 @@ int list_user_groups(int rid, int check)
   if (!check) printf("User is member of %d groups:\n",count);
   
   /* This is the data size */
-  size = get_val_len(hive[H_SAM],nk,"@");
+  size = get_val_len(hive[H_SAM],nk,"@",TPF_VK_EXACT);
   
   /* It should be 4 bytes for each group */
   if (gverbose) printf("Data size %d bytes.\n",size);
@@ -332,7 +338,7 @@ int list_user_groups(int rid, int check)
     printf("list_user_groups(): DEBUG: Size is not 4 * count! May not matter anyway. Continuing..\n");
   }
   
-  m = get_val2buf(hive[H_SAM], NULL, nk, "@", 0);
+  m = get_val2buf(hive[H_SAM], NULL, nk, "@", 0, TPF_VK_EXACT);
   if (!m) {
     printf("list_user_groups(): Could not get value data! Giving up.\n");
     return(0);
@@ -347,7 +353,7 @@ int list_user_groups(int rid, int check)
 
     if (!check) {
       snprintf(g,180,"\\SAM\\Domains\\Builtin\\Aliases\\%08X\\C",grp);
-      c = get_val2buf(hive[H_SAM], NULL, 0, g, 0);
+      c = get_val2buf(hive[H_SAM], NULL, 0, g, 0, TPF_VK_EXACT);
       if (c) {
 	cd = (struct group_C *)&c->data;
 	grpnamoffs = cd->grpname_ofs + 0x34;
@@ -398,7 +404,7 @@ void promote_user(int rid)
     return;
   }
   nk += 4;
-  count = get_val_type(hive[H_SAM],nk,"@");
+  count = get_val_type(hive[H_SAM],nk,"@",TPF_VK);
   if (count == -1) {
     printf("Cannot find value <%s\\@>\n",s);
     return;
@@ -406,15 +412,15 @@ void promote_user(int rid)
   printf("User is member of %d groups.\n",count);
   
   /* This is the data size */
-  size = get_val_len(hive[H_SAM],nk,"@");
+  size = get_val_len(hive[H_SAM],nk,"@",TPF_VK);
   
   /* It should be 4 bytes for each group */
-  printf("Data size %d bytes.\n",size);
   if (size != count * 4) {
+    printf("DEBUG: Data size %d bytes.\n",size);
     printf("DEBUG: Size is not 4 * count! May not matter anyway. Continuing..\n");
   }
   
-  m = get_val2buf(hive[H_SAM], NULL, nk, "@", 0);
+  m = get_val2buf(hive[H_SAM], NULL, nk, "@", 0, TPF_VK);
   if (!m) {
     printf("Could not get value data! Giving up.\n");
     return;
@@ -432,12 +438,12 @@ void promote_user(int rid)
     default: printf(", "); break;
     }
     snprintf(g,180,"\\SAM\\Domains\\Builtin\\Aliases\\%08X\\C",grp);
-    c = get_val2buf(hive[H_SAM], NULL, 0, g, 0);
+    c = get_val2buf(hive[H_SAM], NULL, 0, g, 0, TPF_VK);
     if (c) {
       gcnts = (unsigned int *)&c->data;
       gcnts[0xc]--;
       /* Decrease members counter */
-      put_buf2val(hive[H_SAM], c, 0, g, 0);
+      put_buf2val(hive[H_SAM], c, 0, g, 0, TPF_VK);
     } else {
       printf("Group info for %x not found!\n",grp);
     }
@@ -445,7 +451,7 @@ void promote_user(int rid)
 #if 1
   printf("\nDeleting user memberships\n");
   
-  del_value(hive[H_SAM], nk, "@");
+  del_value(hive[H_SAM], nk, "@", TPF_VK);
   
   printf("Adding into only administrators:\n");
   
@@ -453,14 +459,14 @@ void promote_user(int rid)
     printf("Failed to add @ value to key\n");
   }
 #endif
-  put_buf2val(hive[H_SAM], &admember, nk, "@", 0);
+  put_buf2val(hive[H_SAM], &admember, nk, "@", 0, TPF_VK);
   
   /* Now bumb up administrator groups count */
-  c = get_val2buf(hive[H_SAM], NULL, 0, "\\SAM\\Domains\\Builtin\\Aliases\\00000220\\C", 0);
+  c = get_val2buf(hive[H_SAM], NULL, 0, "\\SAM\\Domains\\Builtin\\Aliases\\00000220\\C", 0, TPF_VK);
   if (!c) printf("Group info for 220 (adm) not found!\n");
   gcnts = (unsigned int *)&c->data;
   gcnts[0xc]++;
-  put_buf2val(hive[H_SAM], c, 0, "\\SAM\\Domains\\Builtin\\Aliases\\00000220\\C", 0);
+  put_buf2val(hive[H_SAM], c, 0, "\\SAM\\Domains\\Builtin\\Aliases\\00000220\\C", 0, TPF_VK);
   
   printf("Promotion DONE!\n");
   
@@ -752,493 +758,14 @@ char *change_pw(char *buf, int rid, int vlen, int stat)
 }
 
 
-/* Here we put our knowledge to use, basic routines to
- * decode and display registry contents almost like a filesystem
- */
-
-/* display (cat) the value,
- * vofs = offset to 'nk' node, paths relative to this (or 0 for root)
- * path = path string to value
- * Does not handle all types yet (does a hexdump instead)
- * MULTI_SZ (multi unicode-string) - only displays first string,
- * but also does a hexdump.
- */
-void cat_vk(struct hive *hdesc, int nkofs, char *path, int dohex)
-{     
-  void *data;
-  int len,i,type;
-  char string[SZ_MAX+1];
-
-  type = get_val_type(hdesc, nkofs, path);
-  if (type == -1) {
-    printf("cat_vk: No such value <%s>\n",path);
-    return;
-  }
-
-  len = get_val_len(hdesc, nkofs, path);
-  if (!len) {
-    printf("cat_vk: Value <%s> has zero length\n",path);
-    return;
-  }
-
-  data = (void *)get_val_data(hdesc, nkofs, path, 0);
-  if (!data) {
-    printf("cat_vk: Value <%s> references NULL-pointer (bad boy!)\n",path);
-    abort();
-    return;
-  }
-
-  printf("Value <%s> of type %s, data length %d [0x%x]\n", path,
-	 (type < REG_MAX ? val_types[type] : "(unknown)"), len, len);
-
-  if (dohex) type = REG_BINARY;
-  switch (type) {
-  case REG_SZ:
-  case REG_EXPAND_SZ:
-  case REG_MULTI_SZ:
-    cheap_uni2ascii(data,string,len);
-    for (i = 0; i < (len>>1)-1; i++) {
-      if (string[i] == 0) string[i] = '\n';
-      if (type == REG_SZ) break;
-    }
-    puts(string);
-    break;
-  case REG_DWORD:
-    printf("0x%08x",*(unsigned short *)data);
-    break;
-  default:
-    printf("Don't know how to handle type yet!\n");
-  case REG_BINARY:
-    hexdump((char *)data, 0, len, 1);
-  }
-  putchar('\n');
-}
-
-/* =================================================================== */
-
-/* Registry editor frontend */
-
-struct cmds {
-  char cmd_str[12];
-  int  cmd_num;
-};
-
-#define MCMD_CD 1
-#define MCMD_LS 2
-#define MCMD_QUIT 3
-#define MCMD_CAT  4
-#define MCMD_STRUCT 5
-#define MCMD_DEBUG 6
-#define MCMD_HELP 7
-#define MCMD_PASSWD 8
-#define MCMD_HIVE 9
-#define MCMD_EDIT 10
-#define MCMD_ALLOC 11
-#define MCMD_FREE 12
-#define MCMD_ADDV 13
-#define MCMD_DELV 14
-#define MCMD_DELVALL 15
-#define MCMD_NEWKEY 16
-#define MCMD_DELKEY 17
-#define MCMD_CATHEX 18
-#define MCMD_RDEL 19
-#define MCMD_CK 20
-
-struct cmds maincmds[] = {
- { "cd" , MCMD_CD } ,
- { "ls" , MCMD_LS } ,
- { "dir", MCMD_LS } ,
- { "q"  , MCMD_QUIT } ,
- { "cat", MCMD_CAT } ,
- { "type",MCMD_CAT } ,
- { "st" , MCMD_STRUCT } ,
- { "pw" , MCMD_PASSWD } ,
- { "passwd", MCMD_PASSWD } ,
- { "debug", MCMD_DEBUG } ,
- { "hive", MCMD_HIVE } ,
- { "ed", MCMD_EDIT } ,
-#if ALLOC_DEBUG
- { "alloc", MCMD_ALLOC } ,
- { "free", MCMD_FREE } ,
-#endif
- { "nv", MCMD_ADDV } ,
- { "dv", MCMD_DELV } ,
- { "delallv", MCMD_DELVALL } ,
- { "nk", MCMD_NEWKEY } ,
- { "dk", MCMD_DELKEY } ,
- { "hex", MCMD_CATHEX } ,
- { "rdel", MCMD_RDEL } ,
- { "ck", MCMD_CK } ,
- { "?", MCMD_HELP } ,
- { "", 0 }
-};
-
-/* Edit value: Invoke whatever is needed to edit it
- * based on its type
- */
-
-void edit_val(struct hive *h, int nkofs, char *path)
-{
-  struct keyval *kv, *newkv;
-  int type,len,n,i,in,go, newsize, d = 0, done, insert = 0;
-  char inbuf[SZ_MAX+4];
-  char origstring[SZ_MAX+4];
-  char *newstring;
-  char *dbuf;
-
-  type = get_val_type(h, nkofs, path);
-  if (type == -1) {
-    printf("Value <%s> not found!\n",path);
-    return;
-  }
-
-  kv = get_val2buf(h, NULL, nkofs, path, type);
-  if (!kv) {
-    printf("Unable to get data of value <%s>\n",path);
-    return;
-  }
-  len = kv->len;
-
-  printf("EDIT: <%s> of type %s with length %d [0x%x]\n", path,
-	 (type < REG_MAX ? val_types[type] : "(unknown)"),
-	 len, len);
-
-  switch(type) {
-  case REG_DWORD:
-    printf("DWORD: Old value %d [0x%x], ", kv->data, kv->data);
-    fmyinput("enter new value (prepend 0x if hex, empty to keep old value)\n-> ",
-	     inbuf, 12);
-    if (*inbuf) {
-      sscanf(inbuf,"%i",&kv->data);
-      d = 1;
-    }
-    printf("DWORD: New value %d [0x%x], ", kv->data, kv->data);
-    break;
-  case REG_SZ:
-  case REG_EXPAND_SZ:
-  case REG_MULTI_SZ:
-    newstring = NULL;
-    dbuf = (char *)&kv->data;
-    cheap_uni2ascii(dbuf,origstring,len);
-    n = 0; i = 0;
-    while (i < (len>>1)-1) {
-      printf("[%2d]: %s\n",n,origstring+i);
-      i += strlen(origstring+i) + 1;
-      n++;
-    }
-
-    printf("\nNow enter new strings, one by one.\n");
-    printf("Enter nothing to keep old.\n");
-    if (type == REG_MULTI_SZ) {
-      printf("'--n' to quit (remove rest of strings)\n");
-      printf("'--i' insert new string at this point\n");
-      printf("'--q' to quit (leaving remaining strings as is)\n");
-      printf("'--Q' to quit and discard all changes\n");
-      printf("'--e' for empty string in this position\n");
-    }
-    n = 0; i = 0; in = 0; go = 0; done = 0;
-
-    /* Now this one is RATHER UGLY :-} */
-
-    while (i < (len>>1)-1 || !done) {
-
-      printf("[%2d]: %s\n",n, insert == 1 ? "[INSERT]" : ((i < (len>>1)-1 ) ? origstring+i : "[NEW]"));
-      if (insert) insert++;
-      if (!go) fmyinput("-> ",inbuf, 500);
-      else *inbuf = 0;
-      if (*inbuf && strcmp("--q", inbuf)) {
-	if (!strcmp("--n", inbuf) || !strcmp("--Q", inbuf)) { /* Zap rest */
-	  i = (len>>1) ; done = 1;
-	} else if (strcmp("--i", inbuf)) {  /* Copy out given string */
-	  if (!strcmp("--e",inbuf)) *inbuf = '\0';
-	  if (newstring) newstring = realloc(newstring, in+strlen(inbuf)+1);
-	  else newstring = malloc(in+strlen(inbuf)+1);
-	  strcpy(newstring+in, inbuf);
-	  in += strlen(inbuf)+1;
-	} else {
-	  insert = 1;
-	}
-      } else {  /* Copy out default string */
-
-	if (newstring) newstring = realloc(newstring, in+strlen(origstring+i)+1);
-	else newstring = malloc(in + strlen(origstring+i) + 1);
-	strcpy(newstring+in, origstring+i);
-	in += strlen(origstring+i)+1;
-
-	if (!strcmp("--q", inbuf)) { 
-	  go = 1; done = 1;
-	  if (!(i < (len>>1)-1 )) {
-	    in--;  /* remove last empty if in NEW-mode */
-	  }
-	}
-      }
-      
-      if (!insert) i += strlen(origstring+i) + 1;
-      if (insert != 1) n++;
-      if (insert == 2) insert = 0;
-      if (type != REG_MULTI_SZ) {
-	i = (len<<1);
-	done = 1;
-      }
-
-    }
-
-    if (strcmp("--Q", inbuf)) {  /* We didn't bail out */
-      if (newstring) newstring = realloc(newstring, in+1);
-      else newstring = malloc(in+1);
-      if (type == REG_MULTI_SZ) {
-	in++;
-	*(newstring+in) = '\0';  /* Must add null termination */
-      }
-      ALLOC(newkv,1,(in<<1)+sizeof(int));
-      newkv->len = in<<1;
-      printf("newkv->len: %d\n",newkv->len);
-      cheap_ascii2uni(newstring, (char *)&(newkv->data), in);
-      
-      d = 1;
-
-      FREE(kv);
-      kv = newkv;
-
-    }
-    break;
-
-  default:
-    printf("Type not handeled (yet), invoking hex editor on data!\n");
-  case REG_BINARY:
-    fmyinput("New length (ENTER to keep same): ",inbuf,90);
-    if (*inbuf) {
-      newsize = atoi(inbuf);
-      ALLOC(newkv,1,newsize+sizeof(int)+4);
-      bzero(newkv,newsize+sizeof(int)+4);
-      memcpy(newkv, kv, (len < newsize) ? (len) : (newsize) +sizeof(int));
-      FREE(kv);
-      kv = newkv;
-      kv->len = newsize;
-    }
-    d = debugit((char *)&kv->data, kv->len);
-    break;
-  }
-
-  if (d) {
-    if (!(put_buf2val(h, kv, nkofs, path, type))) {
-      printf("Failed to set value!?\n");
-    }
-  }
-  FREE(kv);
-}
-
-/* look up command in array
- */
-int parsecmd(char **s, struct cmds *cmd)
-{
-
-  int l = 0;
-
-  while ((*s)[l] && ((*s)[l] != ' ')) {
-    l++;
-  }
-  while (cmd->cmd_num) {
-    if (!strncmp(*s, cmd->cmd_str, l)) {
-      *s += l;
-      return(cmd->cmd_num);
-    }
-    cmd++;
-  }
-  return(0);
-}
-
-/* Simple interactive command-parser
- * Main loop for manually looking through the registry
- */
+/* Registry edit wrapper */
 
 void mainloop(void)
 {
-  struct hive *hdesc;
-  int cdofs, newofs;
-  struct nk_key *cdkey;
-  char inbuf[100], /* whatbuf[100], */ *bp;
-  char path[1000];
-  int l, vkofs, nh;
-  int usehive = 0;
-  struct keyval *kv;
-
-#if ALLOC_DEBUG
-  int pagestart;
-  int freetest;
-#endif
-
-  hdesc = hive[usehive];
-  cdofs = hdesc->rootofs;
-
-  printf("Simple registry editor. ? for help.\n");
-
-  while (1) {
-    cdkey = (struct nk_key *)(hdesc->buffer + cdofs);
-
-    *path = 0;
-    get_abs_path(hdesc,cdofs+4, path, 50);
-
-#if ALLOC_DEBUG
-    pagestart = find_page_start(hdesc,cdofs);
-    printf("find_page_start: 0x%x\n",pagestart);
-    freetest = find_free_blk(hdesc,pagestart,10);
-    printf("find_free_blk: 0x%x\n",freetest);
-#endif
-    if (gverbose) printf("\n[%0x] %s> ",cdofs,path);
-    else printf("\n%s> ",path);
-    l = fmyinput("",inbuf,90);
-    bp = inbuf;
-    skipspace(&bp);
-      
-    if (l > 0 && *bp) {
-      switch(parsecmd(&bp,maincmds)) {
-      case MCMD_HELP:
-	printf("Simple registry editor:\n");
-	printf("hive [<n>] - list loaded hives or switch to hive numer n'\n");
-	printf("cd <key> - change key\nls | dir [<key>] - show subkeys & values,\n");
-        printf("cat | type <value> - show key value\nhex <value> - hexdump of value data\nst [<hexaddr>] - show struct info\n");
-  /* printf("pw | passwd [<hexaddr>] - try the password routine on struct at <hexaddr>\n"); */
-	printf("ck [<keyname>]          - Show keys class data, if it has any\n");
-	printf("nk <keyname> - add key\n");
-	printf("dk <keyname> - delete key (must be empty)\n");
-	printf("ed <value>            - Edit value\n");
-	printf("nv <type> <valuename> - Add value\n");
-	printf("dv <valuename>        - Delete value\n");
-	printf("delallv               - Delete all values in current key\n");
-	printf("rdel <keyname>        - Recursively delete key & subkeys\n");
-	printf("debug - enter buffer hexeditor\nq - quit\n");
-        break;
-
-      case MCMD_DELKEY :
-	bp++;
-	skipspace(&bp);
-        del_key(hdesc, cdofs + 4, bp);
-	break;
-      case MCMD_NEWKEY :
-	bp++;
-	skipspace(&bp);
-        add_key(hdesc, cdofs + 4, bp);
-	break;
-      case MCMD_DELVALL :
-	bp++;
-	skipspace(&bp);
-        del_allvalues(hdesc, cdofs + 4);
-	break;
-      case MCMD_DELV :
-	bp++;
-	skipspace(&bp);
-        del_value(hdesc, cdofs + 4, bp);
-	break;
-      case MCMD_ADDV :
-	bp++;
-	skipspace(&bp);
-	nh = gethex(&bp);
-	skipspace(&bp);
-        add_value(hdesc, cdofs+4, bp, nh);
-	break;
-#ifdef ALLOC_DEBUG
-      case MCMD_FREE :
-	bp++;
-	skipspace(&bp);
-	nh = gethex(&bp);
-        free_block(hdesc, nh);
-	break;
-      case MCMD_ALLOC :
-	bp++;
-	skipspace(&bp);
-	nh = gethex(&bp);
-        alloc_block(hdesc, cdofs+4, nh);
-	break;
-#endif
-      case MCMD_LS :
-	bp++;
-	skipspace(&bp);
-        nk_ls(hdesc, bp, cdofs+4, 0);
-	break;
-      case MCMD_CK :
-	bp++;
-	skipspace(&bp);
-        kv = get_class(hdesc, cdofs+4, bp);
-	if (kv) {
-	  hexdump((char *)&kv->data, 0, kv->len, 1);
-	  FREE(kv);
-	}
-	break;
-      case MCMD_RDEL :
-	bp++;
-	skipspace(&bp);
-        rdel_keys(hdesc, bp, cdofs+4);
-	break;
-      case MCMD_EDIT :
-	bp++;
-	skipspace(&bp);
-        edit_val(hdesc, cdofs+4, bp);
-	break;
-      case MCMD_HIVE :
-	bp++;
-	skipspace(&bp);
-	if (*bp) {
-	  nh = gethex(&bp);
-	  if (nh >= 0 && nh < no_hives) {
-	    usehive = nh;
-	    printf("Switching to hive #%d, named <%s>, size %d [0x%x]\n",
-		   usehive, hive[usehive]->filename,
-		   hive[usehive]->size,
-		   hive[usehive]->size);
-	    hdesc = hive[usehive];
-	    cdofs = hdesc->rootofs;
-	  }
-	} else {
-	  for (nh = 0; nh < no_hives; nh++) {
-	    printf("%c %c %2d %9d 0x%08x <%s>\n", (nh == usehive) ? '*' : ' ',
-		   (hive[nh]->state & HMODE_DIRTY) ? 'D' : ' ',
-		   nh, 
-		   hive[nh]->size,
-		   hive[nh]->size, hive[nh]->filename);
-	  }
-	}
-        break;
-      case MCMD_CD :
-	bp++;
-	skipspace(&bp);
-	newofs = trav_path(hdesc, cdofs+4,bp,0);
-        if (newofs) cdofs = newofs;
-	else printf("Key %s not found!\n",bp);
-	break;
-      case MCMD_CAT:
-	bp++;
-	skipspace(&bp);
-	cat_vk(hdesc,cdofs+4,bp,0);
-	break;
-      case MCMD_CATHEX:
-	bp++;
-	skipspace(&bp);
-	cat_vk(hdesc,cdofs+4,bp,1);
-	break;
-      case MCMD_STRUCT:
-	bp++;
-	skipspace(&bp);
-	vkofs = cdofs;
-	if (*bp) {
-	  vkofs = gethex(&bp);
-	}
-	parse_block(hdesc,vkofs,1);
-	break;
-      case MCMD_DEBUG:
-	if (debugit(hdesc->buffer,hdesc->size)) hdesc->state |= HMODE_DIRTY;
-	break;
-      case MCMD_QUIT:
-        return;
-        break;
-      default:
-	printf("Unknown command: %s\n",bp);
-	break;
-      }
-    }
-  }
+  regedit_interactive(hive, no_hives);
 }
 
+
 /* List users in SAM file
  * pageit - hmm.. forgot this one for this release..
  */
@@ -1265,14 +792,14 @@ int list_users(int pageit)
 
     /* Extract the value out of the username-key, value is RID  */
     snprintf(s,180,"\\SAM\\Domains\\Account\\Users\\Names\\%s\\@",ex.name);
-    rid = get_dword(hive[H_SAM], 0, s);
+    rid = get_dword(hive[H_SAM], 0, s, TPF_VK_EXACT);
     if (rid == 500) strncpy(admuser,ex.name,128); /* Copy out admin-name */
 
     /*    printf("name: %s, rid: %d (0x%0x)\n", ex.name, rid, rid); */
 
     /* Now that we have the RID, build the path to, and get the V-value */
     snprintf(s,180,"\\SAM\\Domains\\Account\\Users\\%08X\\V",rid);
-    v = get_val2buf(hive[H_SAM], NULL, 0, s, REG_BINARY);
+    v = get_val2buf(hive[H_SAM], NULL, 0, s, REG_BINARY, TPF_VK_EXACT);
     if (!v) {
       printf("Cannot find value <%s>\n",s);
       return(1);
@@ -1308,7 +835,7 @@ void find_n_change(char *username)
   if (!rid) { /* Look up username */
     /* Extract the unnamed value out of the username-key, value is RID  */
     snprintf(s,180,"\\SAM\\Domains\\Account\\Users\\Names\\%s\\@",username);
-    rid = get_dword(hive[H_SAM],0,s);
+    rid = get_dword(hive[H_SAM],0,s, TPF_VK_EXACT);
     if (rid == -1) {
       printf("Cannot find value <%s>\n",s);
       return;
@@ -1321,19 +848,19 @@ void find_n_change(char *username)
 
   /* Now that we have the RID, build the path to, and get the V-value */
   snprintf(s,180,"\\SAM\\Domains\\Account\\Users\\%08X\\V",rid);
-  v = get_val2buf(hive[H_SAM], NULL, 0, s, REG_BINARY);
+  v = get_val2buf(hive[H_SAM], NULL, 0, s, REG_BINARY, TPF_VK_EXACT);
   if (!v) {
     printf("Cannot find value <%s>\n",s);
     return;
   }
 
   if (v->len < 0xcc) {
-    printf("Value <%s> is too short (only %ld bytes) to be a SAM user V-struct!\n",
+    printf("Value <%s> is too short (only %d bytes) to be a SAM user V-struct!\n",
 	   s, vkkey->len_data);
   } else {
     change_pw( (char *)&v->data , rid, v->len, 0);
     if (dirty) {
-      if (!(put_buf2val(hive[H_SAM], v, 0, s, REG_BINARY))) {
+      if (!(put_buf2val(hive[H_SAM], v, 0, s, REG_BINARY, TPF_VK_EXACT))) {
 	printf("Failed to write updated <%s> to registry! Password change not completed!\n",s);
       }
     }
@@ -1395,7 +922,7 @@ void handle_syskey(void)
     printf("ERROR: SAM hive not loaded!\n");
     return;
   }
-  samf = get_val2buf(hive[H_SAM], NULL, 0, "\\SAM\\Domains\\Account\\F", REG_BINARY);
+  samf = get_val2buf(hive[H_SAM], NULL, 0, "\\SAM\\Domains\\Account\\F", REG_BINARY, TPF_VK_EXACT);
 
   if (samf && samf->len > 0x70 ) {
     ff = (struct samkeyf *)&samf->data;
@@ -1406,12 +933,12 @@ void handle_syskey(void)
 
   secboot = -1;
   if (H_SYS >= 0) {
-    secboot = get_dword(hive[H_SYS], 0, "\\ControlSet001\\Control\\Lsa\\SecureBoot");
+    secboot = get_dword(hive[H_SYS], 0, "\\ControlSet001\\Control\\Lsa\\SecureBoot", TPF_VK_EXACT);
   }
 
   secmode = -1;
   if (H_SEC >=0) {
-    secpol = get_val2buf(hive[H_SEC], NULL, 0, "\\Policy\\PolSecretEncryptionKey\\@", REG_NONE);
+    secpol = get_val2buf(hive[H_SEC], NULL, 0, "\\Policy\\PolSecretEncryptionKey\\@", REG_NONE, TPF_VK_EXACT);
     if (secpol) {     /* Will not be found in NT 4, take care of that */
       sf = (struct secpoldata *)&secpol->data;
       secmode = sf->syskeymode;
@@ -1419,7 +946,7 @@ void handle_syskey(void)
   }
 
 #ifdef LSADATA
-  lsad = get_val2buf(hive[H_SYS], NULL, 0, "\\ControlSet001\\Control\\Lsa\\Data\\Pattern", REG_BINARY);
+  lsad = get_val2buf(hive[H_SYS], NULL, 0, "\\ControlSet001\\Control\\Lsa\\Data\\Pattern", REG_BINARY, TPF_VK_EXACT);
 
   if (lsad && lsad->len >= 0x38) {
     ld = (struct lsadata *)&lsad->data;
@@ -1463,8 +990,9 @@ void handle_syskey(void)
     printf("if you have lost the key-floppy or passphrase you can turn it off,\n");
     printf("but please read the docs first!!!\n");
     printf("\n** IF YOU DON'T KNOW WHAT SYSKEY IS YOU DO NOT NEED TO SWITCH IT OFF!**\n");
-    printf("NOTE: On WINDOWS 2000 it will not be possible\n");
+    printf("NOTE: On WINDOWS 2000 and XP it will not be possible\n");
     printf("to turn it on again! (and other problems may also show..)\n\n");
+    printf("EXTREME WARNING: Do not try this on Vista, it will go into endless re-boots\n\n");
     printf("NOTE: Disabling syskey will invalidate ALL\n");
     printf("passwords, requiring them to be reset. You should at least reset the\n");
     printf("administrator password using this program, then the rest ought to be\n");
@@ -1480,7 +1008,7 @@ void handle_syskey(void)
 	  ff->syskeyflags1[i] = 0;
 	}
 
-	put_buf2val(hive[H_SAM], samf, 0, "\\SAM\\Domains\\Account\\F", REG_BINARY);
+	put_buf2val(hive[H_SAM], samf, 0, "\\SAM\\Domains\\Account\\F", REG_BINARY, TPF_VK_EXACT);
 
       }
       /* Reset SECURITY infostruct (if any) */
@@ -1488,7 +1016,7 @@ void handle_syskey(void)
 	memset(sf, 0, secpol->len);
 	sf->syskeymode = 0;
 
-	put_buf2val(hive[H_SEC], secpol, 0, "\\Policy\\PolSecretEncryptionKey\\@", REG_BINARY);
+	put_buf2val(hive[H_SEC], secpol, 0, "\\Policy\\PolSecretEncryptionKey\\@", REG_BINARY, TPF_VK_EXACT);
 
       }
 
@@ -1497,14 +1025,14 @@ void handle_syskey(void)
 
 	ld->syskeymode = 0;
 
-	put_buf2val(hive[H_SYS], lsad, 0, "\\ControlSet001\\Control\\Lsa\\Data\\Pattern", REG_BINARY);
+	put_buf2val(hive[H_SYS], lsad, 0, "\\ControlSet001\\Control\\Lsa\\Data\\Pattern", REG_BINARY, TPF_VK_EXACT);
 
       }
 #endif
 
       /* And SYSTEM SecureBoot parameter */
 
-      put_dword(hive[H_SYS], 0, "\\ControlSet001\\Control\\Lsa\\SecureBoot", 0);
+      put_dword(hive[H_SYS], 0, "\\ControlSet001\\Control\\Lsa\\SecureBoot", TPF_VK_EXACT, 0);
 
       dirty = 1;
       syskeyreset = 1;
@@ -1566,11 +1094,11 @@ void recoveryconsole()
     return;
   }
 
-  cmd = get_dword(hive[H_SOF],0,scpath);
-  sec = get_dword(hive[H_SOF],0,slpath);
+  cmd = get_dword(hive[H_SOF],0,scpath,TPF_VK_EXACT);
+  sec = get_dword(hive[H_SOF],0,slpath,TPF_VK_EXACT);
 
   if (cmd == -1 && sec == -1) {
-    printf("\nDid not find registry entries for RecoveryConsole.\n(RecoveryConsole is only in Windows 2000 and XP)\n");
+    printf("\nDid not find registry entries for RecoveryConsole.\n(RecoveryConsole is only in Windows 2000 and XP, not Vista)\n");
     return;
   }
 
@@ -1581,8 +1109,8 @@ void recoveryconsole()
   if (*yn == 'y') {
     cmd ^= 1;
     sec ^= 1;
-    if (!put_dword(hive[0], 0, scpath, cmd)) printf("Update of SET level failed registry edit\n");
-    if (!put_dword(hive[0], 0, slpath, sec)) printf("Update of login level failed registry edit\n");
+    if (!put_dword(hive[0], 0, scpath, TPF_VK_EXACT, cmd)) printf("Update of SET level failed registry edit\n");
+    if (!put_dword(hive[0], 0, slpath, TPF_VK_EXACT, sec)) printf("Update of login level failed registry edit\n");
     printf("Done!\n");
   }
 
